# -*- coding: utf-8 -*-
"""
Created on Fri Feb  9 13:03:04 2018

@author: erwan
"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.misc.arrays import array_allclose
from radis.misc.curve import curve_substract, curve_distance, curve_divide
from radis.spectrum.spectrum import make_up, cast_waveunit
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.widgets import MultiCursor
import numpy as np
from publib import set_style, fix_style

# %% ======================================================================
# External functions
# ----------------
# XXX =====================================================================

def get_diff(s1, s2, var, wunit='default', Iunit='default', medium='default',
             resample=True):
    ''' Get the difference between 2 spectra
    Basically returns w1, I1 - I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.
    
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
    
    var: str
        spectral quantity (ex: 'radiance', 'transmittance'...)
    
    wunit: 'nm', 'cm-1'
        waveunit to compare in
        
    Iunit: str
        If 'default' use s1 unit for variable var
    
    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)
    
    Returns    
    -------
    
    w1, Idiff: array
        difference interpolated on the second range(order?)
    
    
    See Also
    --------
    
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.plot_diff` 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit, 
                                   medium=medium, resample=resample)

    return curve_substract(w1, I1, w2, I2)    # basically w1, I1 - I2 (on same range)


    
def get_ratio(s1, s2, var, wunit='default', Iunit='default', medium='default',
             resample=True):
    ''' Get the ratio between two spectra
    Basically returns w1, I1 / I2 where (w1, I1) and (w2, I2) are the values of
    s1 and s2 for variable var. (w2, I2) is linearly interpolated if needed.
    
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
    
    var: str
        spectral quantity 
    
    wunit: 'nm', 'cm-1'
        waveunit to compare in
        
    Iunit: str
        If 'default' use s1 unit for variable var
    
    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)
    
    
    See Also
    --------
    
    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_distance`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.plot_diff` 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    
    '''

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit, 
                                   medium=medium, resample=resample)

    return curve_divide(w1, I1, w2, I2)    # basically w1, I1 - I2 (on same range)

def get_distance(s1, s2, var, wunit='default', Iunit='default', medium='default',
             resample=True):
    ''' Get the Euclidian distance between 2 spectra
    
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
    
    var: str
        spectral quantity 
    
    wunit: 'nm', 'cm-1'
        waveunit to compare in
        
    Iunit: str
        If 'default' use s1 unit for variable var
    
    medium: 'air', 'vacuum', default'
        propagating medium to compare in (if in wavelength)
    
    
    See Also
    --------
    
    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`,  
    :func:`~radis.spectrum.compare.get_residual`,
    :func:`~radis.spectrum.compare.plot_diff` 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    
    '''

    w1, I1, w2, I2 = _get_defaults(s1, s2, var=var, wunit=wunit, Iunit=Iunit, 
                                   medium=medium, resample=resample)

    return curve_distance(w1, I1, w2, I2, discard_out_of_bounds=True)    # euclidian distance from w1, I1

def get_residual(s1, s2, var, ignore_nan=False):
    ''' Returns integral of the difference between two spectra s1 and s2, 
    relatively to the integral of spectrum s1 
    
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
    
    var: str
        spectral quantity
    
    Other Parameters
    ----------------
    
    ignore_nan: boolean
        if True, ignore nan in the difference between s1 and s2 (ex: out of bound)
        when calculating residual. Default False. Note: get_residual will still 
        fail if there are nan in initial Spectrum. 
        
    Notes
    -----
    
    For I1, I2, the values of 'var' in s1 and s2, respectively, residual
    is calculated as:
    
    >>> res = trapz(I2-I1, w1) / trapz(I1, w1)

    0 values for I1 yield nans except if I2 = I1 = 0
    
    when s1 and s2 dont have the size wavespace range, they are automatically
    resampled through get_diff on 's1' range 

    
    See Also
    --------
    
    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.plot_diff` 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    '''

    w, I = s1.get(var)
    # mask for 0
    wdiff, dI = get_diff(s1, s2, var, resample=True)
    
    if ignore_nan:
        b = np.isnan(dI)
        wdiff, dI = wdiff[~b], dI[~b]
    
    return np.abs(np.trapz(dI, wdiff) / np.trapz(I, w))

#    b1 = (I_avg == 0)
#    b2 = (dI == 0)
    
#    with catch_warnings():
#        filterwarnings('ignore', 'invalid value encountered in true_divide')
#        res = dI/I_avg
    
#    res[b1*b2] = 0             # if both I_avg and dI = 0, solve nan as 0     
    
#    return np.sqrt((res**2).sum())/len(res)
#    return res.mean()

def _get_defaults(s1, s2, var, wunit='default', Iunit='default', medium='default',
             resample=True):
    ''' See get_distance, get_diff '''

    # Check inputs, get defaults
    # ----
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum. Cant plot')        
    # Format units    
    if wunit == 'default':
        wunit = s1.get_waveunit()
    wunit = cast_waveunit(wunit)
    if medium == 'default':
        medium = s1.conditions.get('medium', None)

    # Get data
    # ----
    w1,I1 = s1.get(var, wunit=wunit, Iunit=Iunit, medium=medium)
    w2,I2 = s2.get(var, wunit=wunit, Iunit=Iunit, medium=medium)
    
    if not resample:
        if not array_allclose(w1, w2):
            raise AssertionError('Wavespace are not the same: use Spectrum.resample()')

    return w1, I1, w2, I2

def plot_diff(s1, s2, var=None, wunit='default', Iunit='default', medium='default',
              resample=True, method='diff',  show_points=False,
              label1 = None, label2 = None, figsize=None, title=None, nfig=None,
              normalize=False, verbose=True):
    ''' Plot two spectra, and the difference between them
    
    
    Parameters    
    ----------
    
    s1, s2: Spectrum objects
    
    var: str, or None
        spectral quantity to plot (ex: 'abscoeff'). If None, plot the first one 
        in the Spectrum from 'radiance', 'radiance_noslit', 'transmittance', etc.
        
    wunit: 'default', 'nm', 'cm-1'
        if 'default', use first spectrum wunit
    
    Iunit: str
        if 'default', use first spectrum unit
        
    medium: 'air', 'vacuum', 'default'
        if 'default', use first spectrum propagating medium
        
    method: 'distance', 'diff', 'ratio'
        If 'diff', plot difference at same wavespace position. 
        If 'distance', plot Euclidian distance (note that units are meaningless then)
        If 'ratio', plot ratio of two spectra
        Default 'diff'.
        
        Warning: with 'distance', calculation scales as ~N^2 with N the number
        of points in a spectrum (against ~N with 'diff'). This can quickly 
        override all memory.
        
    normalize: bool
        Normalize the spectra to be ploted 
        
    Other Parameters
    ----------------
    
    show_points: boolean
        if True, make all points appear with 'o'
    
    label1, label2: str
        curve names
        
    figsize
        figure size
        
    nfig: int, str
        figure number of name
        
    title: str
        title

    verbose: boolean
        if True, plot stuff such as rescale ratio in normalize mode. Default True


    Examples
    --------

    >>> Punit = 'mW/cm2/sr'
    >>> fig, axes = plot_diff(s10, s50, figsize=(18,6),
    >>>       label1='brd 10 cm-1, P={0:.2f} {1}'.format(s10.get_power(unit=Punit),Punit),
    >>>       label2='brd 50 cm-1, P={0:.2f} {1}'.format(s50.get_power(unit=Punit),Punit)
    >>>       )
    
    
    
    See Also
    --------
    
    :func:`~radis.spectrum.compare.get_diff`, 
    :func:`~radis.spectrum.compare.get_ratio`, 
    :func:`~radis.spectrum.compare.get_distance`, 
    :func:`~radis.spectrum.compare.get_residual`, 
    :meth:`~radis.spectrum.spectrum.compare_with` 
    
    '''
    
    # Get defaults
    # ---
    if var is None:    # if nothing is defined, try these first:
        params = s1.get_vars()
        if 'radiance' in params:
            var = 'radiance'
        elif 'radiance_noslit' in params:
            var = 'radiance_noslit'
        elif 'transmittance' in params:
            var = 'transmittance'
        elif 'transmittance_noslit' in params:
            var = 'transmittance_noslit'
        else:
            # or plot the first variable we find
            var = list(params)[0]
            if var.replace('_noslit', '') in params:
                var = var.replace('_noslit', '')
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum. Cant plot')
    if wunit == 'default':
        wunit = s1.get_waveunit()
    if medium == 'default':
        medium = s1.conditions.get('medium', None)

    # Get data
    # ----
    if method == 'distance':
        wdiff, Idiff = get_distance(s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
    elif method == 'diff':
        wdiff, Idiff = get_diff(s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
    elif method == 'ratio':
        wdiff, Idiff = get_ratio(s1, s2, var=var, wunit=wunit, Iunit=Iunit, medium=medium)
    else:
        raise ValueError('Unknown comparison method: {0}'.format(method))
    
    # Plot 
    # ----
    
    # Format units
    wunit = cast_waveunit(wunit)
    if wunit == 'cm-1':
        xlabel='Wavenumber (cm-1)'
    elif wunit == 'nm':
        xlabel='Wavelength (nm)'
        
    # Init figure
    set_style('origin')
    fig = plt.figure(num=nfig, figsize=figsize)
    gs = gridspec.GridSpec(2, 1, height_ratios=[3, 1])
    ax0 = plt.subplot(gs[0])
    ax1 = plt.subplot(gs[1])
    ax1.get_shared_x_axes().join(ax0, ax1)

    # Plotting style
    if show_points:
        style = '-o'
    else:        
        style = '-'

    # Get labels and names
    if label1 is None:
        label1 = s1.get_name()
    if label2 is None:
        label2 = s2.get_name()
    # Plot compared spectra
    if normalize:
        w1, I1 = s1.get(var, wunit, Iunit, medium)
        w2, I2 = s2.get(var, wunit, Iunit, medium)
        if verbose: print(('Rescale factor: '+str(np.max(I1)/np.max(I2))))
        ax0.plot(w1, I1/np.max(I1), ls=style, color='k', lw=3, label=label1)
        ax0.plot(w2, I2/np.max(I2), ls=style, color='r', lw=1, label=label2)
    else:
        ax0.plot(*s1.get(var, wunit, Iunit, medium), ls=style, color='k', lw=3, label=label1)
        ax0.plot(*s2.get(var, wunit, Iunit, medium), ls=style, color='r', lw=1, label=label2)

    Iunit = make_up(Iunit)  # cosmetic changes 
    
    ax0.tick_params(labelbottom='off')
    if label1 is not None or label2 is not None:
        ax0.legend(loc='best')
    
    # Start to 0
    if var in ['radiance_noslit', 'radiance', 'abscoeff', 'absorbance']:
        ax0.set_ylim(bottom=0)

    # plot difference (sorted)
    b = np.argsort(wdiff)
    ax1.plot(wdiff[b], Idiff[b], style, color='k', lw=1)
    
    if method == 'diff':
        fig.text(0.09,0.38,'diff')
    elif method == 'distance':
        fig.text(0.09,0.38,'distance')
    elif method == 'ratio':
        fig.text(0.09,0.38,'ratio')
        
    # Write labels
    ax1.set_xlabel(make_up(xlabel))
    if normalize:
       fig.text(0.02, 0.5, 'Arb. Units',
                 va='center', rotation='vertical')
    else:
        fig.text(0.02, 0.5, ('{0} ({1})'.format(make_up(var), Iunit)),
                 va='center', rotation='vertical')
        
    # Set limits
    if method == 'diff':
        # symmetrize error scale:
        ymin, ymax = ax1.get_ylim()
        ymax = max(abs(ymin), abs(ymax))
        ax1.set_ylim(-ymax, ymax)
    elif method == 'distance':
        ax1.set_ylim(bottom=0)
    elif method == 'ratio':
        # auto-zoom on min, max, but discard first decile (case of spikes / divergences)
        Idiff_s = np.sort(Idiff)
        ax1.set_ylim(bottom=0, top=Idiff_s[len(Idiff_s)//10]+Idiff_s[-len(Idiff_s)//10])
    

    if title:
        fig.suptitle(title)

    # Fix format
    fix_style('origin', ax=ax0)
    fix_style('origin', ax=ax1)
    plt.tight_layout()
    if title:
        plt.subplots_adjust(left=0.15, top=0.92)
    else:
        plt.subplots_adjust(left=0.15)
        
    # Add cursors
    fig.cursors = MultiCursor(fig.canvas, (ax0, ax1), 
                         color='r', lw=1, alpha=0.2, horizOn=False, 
                         vertOn=True)
    
    return fig, [ax0, ax1]
''' Return the average distance between two spectra.
    It's important to note that if averageDistance(s1, s2)==0 then s1 = s2
    

   .. math::

      \\sqrt{\\sum {(u_i-v_i)^2 / V[x_i]}}.
'''

def averageDistance(s1, s2, var='radiance'):
    ''' Return the average distance between two spectra.
    It's important to note for fitting that if averageDistance(s1, s2)==0 
    then s1 = s2
    
    Parameters    
    ----------
            
    s1, s2: Spectrum objects
        spectra to be compared
    
    var: str, optional
        spectral quantity (ex: 'radiance', 'transmittance'...)
    
    Returns    
    -------
    
    distance: float
        Average distance as in the following equation:
        
        .. math::

            dist = \\frac{\\sqrt{\\sum_i {(s1_i-s2_i)^2}}}{N}.
    
    '''
    distArray = get_diff(s1, s2, var)
    distArray_Y2 = distArray[1]*distArray[1]
    N = np.size(distArray[1])
    distance = np.sqrt(np.sum(distArray_Y2))/N
    return distance


if __name__ == '__main__':
    
    from radis.test.spectrum.test_compare import _run_testcases
    print(('Test compare.py: ', _run_testcases(verbose=True)))
    


# -*- coding: utf-8 -*-
"""

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have 'fast' in their name)

>>> pytest -k fast

"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.spectrum import Spectrum, calculated_spectrum
from radis.phys.convert import nm2cm
#from radis.misc.utils import DatabankNotFound
#from radis.test.utils import IgnoreMissingDatabase, build_test_databases
import numpy as np
from numpy import allclose, linspace
import os
from os.path import basename, exists

fig_prefix = basename(__file__)+': '

# %% Test routines

def test_spectrum_get_methods(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Test all spectrum methods on a Spectrum generated in Specair '''
    
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    from radis.tools.slit import get_FWHM


    if plot and close_plots:
        import matplotlib.pyplot as plt
        plt.close('all')

    s = load_spec(getTestFile('N2C_specair_380nm.spec'))
    
    # general methods
    if verbose: print(s)
    dir(s)
    
    # access properties
    assert s.get_name() == 'N2C_specair_380nm'
    assert all(s.get_radiance_noslit(Iunit='W/m2/sr/nm') == 
               s.get('radiance_noslit', Iunit='W/m2/sr/nm')[1])
    assert all(nm2cm(s.get_wavelength(medium='vacuum')) == 
               s.get_wavenumber())
    assert s.get_power(unit='W/cm2/sr') == 2631.6288408588148
    assert s.get_waveunit() == 'nm'
    assert s.get_power(unit='W/cm2/sr') == s.get_integral('radiance_noslit', Iunit='W/cm2/sr/nm')
    assert s.get_conditions()['Tgas'] == 1500
    assert s.get_medium() == 'air'
    assert len(s.get_vars()) == 2
    assert s.is_at_equilibrium() == False
    assert s.is_optically_thin() == False
    
    # Check applied slit has the correct width
    s.apply_slit(0.5)
    wslit, Islit = s.get_slit()
    wstep = np.diff(wslit)[0]
    assert np.isclose(get_FWHM(*s.get_slit()), 0.5, atol=1.1*wstep)
    
    if plot:
        s.plot_slit()
    
    if verbose:
        print('Tested Spectrum methods:')
        print('...print(Spectrum)')
        print('.. get_name()')
        print('.. get_radiance_noslit() vs get()')
        print('.. get_wavelength() vs get_wavenumber')
        print('.. get_power()')
        print('.. get_waveunit()')
        print('.. get_power() vs get_integral()')
        print('.. get_conditions()')
        print('.. get_vars()')
        print('.. is_at_equilibrium()')
        print('.. is_optically_thin()')
        print('.. get_slit()')

def test_copy(verbose=True, *args, **kwargs):
    ''' Test that a Spectrum is correctly copied 
    
    We compare a Spectrum that has:
    - all available spectral quantities
    - a slit 
    - many calculation conditions
    - no populations
    - no lines 
    '''
    
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s.update()
    s.apply_slit(1.5)
    s2 = s.copy()
    
    assert s == s2
    assert s is not s2
    
    if verbose:
        print('Tested that s2 == s after Spectrum copy')
    
def test_populations(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Test that populations in a Spectrum are correctly read 
    '''
    
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
    

    if plot and close_plots:
        import matplotlib.pyplot as plt
        plt.close('all')

    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    import pytest
    
    # get a spectrum
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    
    # Get all populations
    pops = s.get_populations(molecule='CO', isotope=1)
    assert np.isclose(pops['Ia'], 0.986544)
    
    # Get vib levels
    df_vib = s.get_vib_levels()
    assert len(df_vib) == 49
    
    # Get rovib levels (dont exist: should fail!)
    with pytest.raises(KeyError):  # expected behavior
        s.get_rovib_levels()
        
    if plot:
        s.plot_populations()
    
    if verbose:
        print('test_populations: OK')
    
def test_store_functions(verbose=True, *args, **kwargs):
    ''' Test some store / retrieve functions '''
    
    from radis.spectrum.spectrum import transmittance_spectrum
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    
    temp_file = 'test_radis_tempfile_transmittance.txt'
    assert not exists(temp_file)
         
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    s.update()
    try:
        s.savetxt(temp_file, 'transmittance_noslit', wunit='nm', medium='vacuum')
        w, T = np.loadtxt(temp_file).T
    finally:
        os.remove(temp_file)

    s2 = transmittance_spectrum(w, T, wunit='nm', conditions={'medium':'vacuum'})       
    assert s.compare_with(s2, spectra_only='transmittance_noslit', plot=False)
    
    return True

def test_intensity_conversion__fast(verbose=True, *args, **kwargs):
    ''' Test conversion of intensity cm-1 works:
        
    - conversion of mW/sr/cm2/nm -> mW/sr/cm2/cm-1
    
    '''

    from radis import planck, planck_wn

    w_nm = linspace(300, 3000)
    w_cm = nm2cm(w_nm)
    I_nm = planck(w_nm, T=6000, unit='mW/sr/cm2/nm')

    s = calculated_spectrum(w_nm, I_nm, wunit='nm', Iunit='mW/sr/cm2/nm', 
                            conditions={'medium':'vacuum'})
    
    # mW/sr/cm2/nm -> mW/sr/cm2/cm-1
    w, I = s.get('radiance_noslit', Iunit='mW/sr/cm2/cm_1')
    I_cm = planck_wn(w_cm, T=6000, unit='mW/sr/cm2/cm_1')
    assert allclose(I_cm, I, rtol=1e-3)
    
#def test_emissivity_conversion__fast(verbose=True, *args, **kwargs):
#    ''' Test conversion of intensity cm-1 works:
#        
#    - conversion of mW/sr/cm2/nm -> mW/sr/cm2/cm-1
#    
#    '''
#    
#    from radis.phys.units import convert_emi2nm
#    from radis.test.utils import getTestFile
#    from radis.tools.database import load_spec
#    
#    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
#    s.update()
#
#    # mW/sr/cm3/nm -> mW/sr/cm3/cm-1
#    
#    w, I = s.get('emissivity_noslit', Iunit='mW/sr/cm3/cm_1')
#    I_cm = convert_emi2nm(I, 'mW/sr/cm3/cm_1', 'mW/sr/m3/µm')
#    
    
# TODO: finish implementing emissivity_conversino above


def test_rescaling_function__fast(verbose=True, *args, **kwargs):
    ''' Test rescaling functions '''

    from radis.test.utils import getTestFile

    s = Spectrum.from_txt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000.txt'),
                          quantity='radiance_noslit', 
                          waveunit='nm',
                          unit='mW/cm2/sr/µm',   # Specair units: mW/cm2/sr/µm
                          conditions={'Tvib':3000, 'Trot':1200,
                                        'path_length':1,  # arbitrary
                                        'medium':'air',
                                        },
                          populations={'molecules':{'N2C':1e13}}, # arbitrary
                                                            # (just an example)
                          )
    s.update(optically_thin=True)
    s.rescale_path_length(10)

    assert np.isclose(s.get_radiance_noslit(Iunit='mW/cm2/sr/nm')[0], 352.57305783248)

def test_resampling_function__fast(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Test resampling functions 
    
    Get a Spectrum calculated in cm-1, then resample on a smaller range in cm-1, 
    and in approximately the same range (but in nm). Check that all 3 overlap 
    '''
# %%
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    from radis.spectrum import get_residual

    if plot and close_plots:
        import matplotlib.pyplot as plt
        plt.close('all')

    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    s.name = 'original'
    s2 = s.copy()
    s2b = s.copy()
    s3 = s.copy()
    s2.resample(np.linspace(4500, 4700, 10000), unit='nm', medium='vacuum')
    s2b.resample(np.linspace(4500, 4700, 10000), unit='nm', medium='air')
    s3.resample(np.linspace(2127.2, 2227.7, 10000), unit='cm-1')
    s2.name = 'resampled in nm (vacuum)'
    s2b.name = 'resampled in nm (air)'
    s3.name = 'resampled in cm-1'
    
    s.compare_with(s2, plot=plot, title='Residual: {0:.2g}'.format(
            get_residual(s, s2, 'abscoeff', ignore_nan=True)))
    s.compare_with(s2b, plot=plot, title='Residual: {0:.2g}'.format(
            get_residual(s, s2b, 'abscoeff', ignore_nan=True)))
    s.compare_with(s3, plot=plot, title='Residual: {0:.2g}'.format(
            get_residual(s, s3, 'abscoeff', ignore_nan=True)))
    
    assert get_residual(s, s2, 'abscoeff', ignore_nan=True) < 1e-4
    assert get_residual(s, s2b, 'abscoeff', ignore_nan=True) < 1e-3
    assert get_residual(s, s3, 'abscoeff', ignore_nan=True) < 1e-5
    
# %%

def _run_testcases(plot=True, close_plots=False, verbose=True, debug=False, warnings=True, *args, **kwargs):
    ''' Test procedures

    Input
    ------

    debug: boolean
        swamps the console namespace with local variables. Default False

    '''
    
    # Test all Spectrum methods 
    # -------------------------
    test_spectrum_get_methods(debug=debug, verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs)
    test_copy(verbose=verbose, *args, **kwargs)
    test_populations(verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs)
    test_store_functions(verbose=verbose, *args, **kwargs)
    
    
    # Test populations
    # ----------
#    test_populations__fast(verbose=verbose, *args, **kwargs)

    # Test conversion of intensity cm-1 works
    # -------------
    test_intensity_conversion__fast(debug=debug, verbose=verbose, *args, **kwargs)

    # Test updating / rescaling functions (no self absorption)
    # ---------
    test_rescaling_function__fast(debug=debug, *args, **kwargs)
    test_resampling_function__fast(debug=debug, plot=plot, close_plots=close_plots, *args, **kwargs)
#    test_rescaling_path_length__fast(plot=plot, verbose=verbose, debug=debug,
#                                     warnings=warnings, *args, **kwargs)
#    test_rescaling_mole_fraction__fast(plot=plot, verbose=verbose, debug=debug,
#                                     warnings=warnings, *args, **kwargs)
    
    # Test propagating medium
#    test_medium__fast(plot=plot, verbose=verbose, debug=debug,
#                           warnings=warnings, *args, **kwargs)

    return True


if __name__ == '__main__':
    print(('Test spectrum: ', _run_testcases(debug=False, close_plots=False)))



# -*- coding: utf-8 -*-
"""
Created on Sun Jan  3 17:52:04 2016

@author: Erwan

Debug functions
"""

from __future__ import print_function, absolute_import, division, unicode_literals

# %%
#==============================================================================
# Debug functions
#==============================================================================

def printdbg(*args, **kwargs):
    ''' Function that prints only in debug mode. change this at runtime with 
     >>> radis.debug = True
     
     Examples
     --------
     
     Embed this print in a if __debug__ statement::
     
         if __debug__: printdbg(...)
         
     so that printdbg are removed by the Python preprocessor when running in 
     optimize mode::
     
         python -O *.py
     
     Notes
     -----
     
     Todo: 
     
     - no idea how this performs in multiprocessing mode
     '''
     
    from radis import DEBUG_MODE
    
    if DEBUG_MODE:
        print('DEBUG:',*args, **kwargs)

def export(var=locals()):
    ''' Export local variables. Useful for debugging 

    Debugging inside a function may be tedious because you can't access the 
    local variables. One of the option is to use the ipython magic::

        %debug    

    Or the pdb equivalent::

        import pdb
        pdb.pm()

    Another option is to insert this export() call in the troubled function, 
    before the exception occurs.


    Examples
    --------
    
        debug_export(locals())

    Note: you can also use  'globals().update(locals())' directly in your 
    function to debug

    Note
    ----
    
    - seems not to work for functions nested in functions
    - 01/05 : doesn't seem to work at all.. @Erwan  


    '''

    globals().update(var)

    return
#
# def get_locals(func):
#    ''' Initially from :
#    http://stackoverflow.com/questions/20577806/keep-function-namespace-alive-for-debugging-in-ipython
#
#    '''
#
#    def wrap(*args, **kw):
#        sys.settrace(tracefunc)
#        try:
#            res = func(*args, **kw)
#        finally:
#            sys.settrace(None)
#        return res
#
#    def tracefunc(frame, event, arg):
#        if event == "return":
#            if frame.f_code is func.func_code:
#                wrap.last_res = frame.f_locals
#        return tracefunc
#
#    return wrap




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 20 23:44:47 2018
Handy functions for manipulation of spectra
@author: minou
"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.misc.curve import curve_substract, curve_add
from radis import get_diff, Spectrum

def substract(s1, s2, var='radiance', wunit='nm', Iunit='default', 
              resample=True, name='default'):
    '''Substract s2 to s1

    Parameters    
    ----------
    s1, s2: Spectrum objects
        The spectra to substract.
    var: str, optional
        Spectral quantity (ex: 'radiance', 'transmittance'...).
    wunit: str, optional 
        Wave unit ('nm', 'cm-1', ...) to use for output. If default, s1 wunit is taken.        
    Iunit: str
        If 'default' s1 unit is used for variable var.
    name: str, optional
        If not given will be s1 and s2 names separated by '+'.
    Returns    
    ----------
    sub : Spectrum object where intensities of s1 and s2 are substracted
    
    Note    
    ----------
    Wavelength of s1 is taken as reference.
    '''
    if s1.get_conditions()['medium'] != s2.get_conditions()['medium']:
        raise ValueError('Not the same medium for s1 and s2')
        # Check inputs, get defaults
    # ----
    if Iunit == 'default':
        try:
            Iunit = s1.units[var]
        except KeyError:  # unit not defined in dictionary
            raise KeyError('Iunit not defined in spectrum. Cant plot') 
            
    w1, I3 = get_diff(s1, s2, var=var, wunit=wunit, Iunit=Iunit, 
                                  resample=resample)
    if name != 'default':
        name = s1.get_name()+'+'+s2.get_name()
    sub = Spectrum.from_array(w1, I3, var, 
                               waveunit=wunit, 
                               unit=Iunit,
                               conditions={'medium' : s1.conditions['medium'], 'waveunit':wunit}, 
                               name=name)
    return sub



# -*- coding: utf-8 -*-
"""
Test that line survey works
"""

from __future__ import print_function, absolute_import, division, unicode_literals
#from radis.misc.utils import DatabankNotFound
from radis.test.utils import getTestFile
from radis.tools.database import load_spec, SpecDatabase
import os
from os.path import exists, dirname, basename

def test_database_functions__fast(verbose=True, plot=True, close_plots=True, warnings=True, *args, **kwargs):
    ''' Test SpecDatabase functions '''
    
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()
    if close_plots: plt.close('all')
    
    import pytest
    
    db = SpecDatabase(dirname(getTestFile('.')))

    # Database visualisation methods    
    if verbose:
        print('{0} items in test database: {1}'.format(len(db), db.see(['Tvib', 'Trot'])))
    if plot:
        db.plot('Tvib', 'Trot')

    # Database get methods
    db.get_closest(Tgas=1300, path_length=1)
    s = db.get_unique(Tgas=1500, path_length=0.01, mole_fraction=0.5)  # there should be one only
                                                      # ... note that this is just so we can test 
                                                      # ... get_unique(). If we were to add new
                                                      # ... test cases with matching conditions
                                                      # ... let's add more criteria to keep it unique
    match = db.get(**s.conditions)
#    assert len(match) == 1 
    # TODO: not working in Python 2.7 yet
    
    # Database add method
    s2 = s.copy()
    s2.conditions['Tgas'] = 0  # make it unique (for testing)
    l = db.add(s2, add_info=['Tvib', 'Trot'], discard=[], compress=True, if_exists_then='increment')
    assert exists(l)
    try:
        assert s2 in db
        # .. ensures that you cant add it twice
        with pytest.raises(ValueError):
            db.add(s2, add_info=['Tvib', 'Trot'])
    finally:
        os.remove(l)
    db.update(force_reload=True)            # update database 
    assert s2 not in db
    
def test_plot_spec(plot=True, close_plots=True, verbose=True, *args, **kwargs):
    
    from radis.tools.database import plot_spec
    from radis.test.utils import getTestFile
    
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()
    if close_plots: plt.close('all')
    
    if plot:
        plot_spec(getTestFile('N2C_specair_380nm.spec'))
        

def _run_testcases(plot=True, close_plots=False, verbose=True, *args, **kwargs):

    test_database_functions__fast(plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs)
    test_plot_spec(plot=plot, close_plots=close_plots, ssverbose=verbose, *args, **kwargs)

    return True

if __name__ == '__main__':

    print('Testing line survey functions:', _run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 13 21:58:47 2017

@author: erwan

Formula to calculate most spectral quantities

Are stored here all equations that need to be used by both the factory (to 
generate the spectrum) or the spectrum class (to recompute some quantities
in post processing)

"""

from __future__ import absolute_import
from radis.phys.convert import cm2nm
from radis.phys.blackbody import planck

# ------- Get radiance from emissivity

def calc_radiance(wavenumber, emissivity, Tgas, unit='mW/sr/cm2/nm'):
    ''' Derive radiance (mW/cm2/sr/nm) from the emissivity


    Returns
    -------

    radiance: mW/sr/cm2/nm
    '''

    radiance = emissivity * planck(cm2nm(wavenumber), Tgas, unit=unit)

    return radiance




# -*- coding: utf-8 -*-
"""
Created on Fri Apr 14 12:06:46 2017

@author: erwan
"""

from __future__ import print_function, absolute_import, division, unicode_literals

import numpy as np
from radis.io.hitran import get_molecule
from radis.io.hitran  import columns_2004 as hitrancolumns
from radis.io.cdsd import columns_hitemp as cdsdcolumns
from radis.io.cdsd import columns_4000 as cdsd4000columns
from radis.phys.convert import cm2nm
from radis.phys.air import vacuum2air
from radis.phys.constants import k_b
from radis.misc.basics import is_float
from radis.misc.utils import NotInstalled
from warnings import warn
try:
    import plotly.offline as py
    import plotly.graph_objs as go
except ImportError:
    py = NotInstalled(name='plotly',info='LineSurvey requires Plotly. Please install it manually')
    go = NotInstalled(name='plotly',info='LineSurvey requires Plotly. Please install it manually')

def LineSurvey(spec, overlay=None, wunit='cm-1',Iunit='hitran', medium='air', 
               cutoff=None, plot='S', lineinfo=['int', 'A'], barwidth=0.07, 
               yscale='log', 
               display=True, filename='line-survey.html',
               xunit=None, yunit=None # deprecated
               ):    
    ''' Plot Line Survey (all linestrengths above cutoff criteria) in Plotly (html)

    
    Parameters    
    ----------

    spec: Spectrum 
        result from SpectrumFactory calculation (see spectrum.py)

    overlay: tuple (w, I, [name], [units]), or list or tuples
        plot (w, I) on a secondary axis. Useful to compare linestrength with
        calculated / measured data 
    
    xunit: 'nm', 'cm-1'
        wavelength / wavenumber units

    Iunit: `hitran`, `splot` 
        Linestrength output units:
            
        - `hitran`: (cm-1/(molecule/cm-2))
        - `splot`: (cm-1/atm)   (Spectraplot units [2]_)
        
        Note: if not None, cutoff criteria is applied in this unit.
        Not used if plot is not 'S'

    medium: 'air', 'vacuum'
        show wavelength either in air or vacuum. Default air
        
    plot: str
        what to plot. Default 'S' (scaled line intensity). But it can be 
        any key in the lines, such as population ('nu'), or Einstein coefficient ('Aul')

    lineinfo: list, or 'all'
        extra line information to plot. Should be a column name in the databank
        (s.lines). For instance: 'int', 'selbrd', etc... Default ['int']

    Other Parameters
    ----------------
    
    display: boolean
        if True, open the image in a browser. Default True. 

    filename: str
        filename to save .html 

    yscale: 'log', 'linear'
        Default 'log'
        
    Returns
    -------
    
    Plot in Plotly. See Output in [1]_
    
        
    Examples
    --------
        
    An example using the :class:`~radis.lbl.factory.SpectrumFactory` to generate a spectrum::
    
        from radis import SpectrumFactory
        sf = SpectrumFactory(
                             wavenum_min=2380,
                             wavenum_max=2400,
                             mole_fraction=400e-6,
                             path_length=100,  # cm
                             isotope=[1],
                             db_use_cached=True) 
        sf.load_databank('HITRAN-CO2-TEST')
        s = sf.eq_spectrum(Tgas=1500)
        s.apply_slit(0.5)
        s.line_survey(overlay='radiance_noslit', barwidth=0.01)

    References
    ----------
    
    .. [1] `RADIS Online Documentation (LineSurvey) <https://radis.readthedocs.io/en/latest/tools/line_survey.html>`__

    .. [2] `SpectraPlot <http://www.spectraplot.com/survey>`__


    '''
    
    # Check inputs
    if xunit is not None:
        warn(DeprecationWarning('xunit replaced with wunit'))
        wunit = xunit
    if yunit is not None:
        warn(DeprecationWarning('yunit replaced with Iunit'))
        Iunit = yunit
    assert yscale in ['log', 'linear']
   
    try:
       spec.lines
    except AttributeError:
        raise AttributeError('Spectrum has no `lines` property. Cant run line survey')

    T = spec.conditions['Tgas']
    P = spec.conditions['pressure_mbar']/1000 # bar1013.25 # atm
    Xi = spec.conditions['mole_fraction']
    sp = spec.lines.copy()
    dbformat = spec.conditions['dbformat']
    
    if not plot in list(sp.keys()):
        raise KeyError('Key {0} is not in line database: {1}'.format(plot, list(sp.keys())))

    def hitran2splot(S):
        ''' convert Linestrength in HITRAN units (cm-1/(molecules.cm-2)) to
        SpectraPlot units (cm-2/atm) '''

        return S/(k_b*T)/10

    # Parsers to get units and more details
    if dbformat == 'hitran':   
        columndescriptor = hitrancolumns 
    elif dbformat == 'cdsd':
        columndescriptor = cdsdcolumns
    elif dbformat == 'cdsd4000':
        columndescriptor = cdsd4000columns
    
    # Apply cutoff, get ylabel
    if plot == 'S':
        if Iunit == 'hitran':
            if cutoff is None: cutoff = spec.conditions['cutoff']
            sp = sp[(sp.S>cutoff)]
            Iunit_str = 'cm-1/(molecule/cm-2)'
        elif Iunit == 'splot':
            if cutoff is None:
                cutoff = spec.conditions['cutoff']
                sp = sp[(sp.S>cutoff)] # if None, use default cutoff expressed in Hitran units
            else:
                sp['S'] = hitran2splot(sp.S)
                sp = sp[(sp.S>cutoff)]
            Iunit_str=  'cm-1/atm'
        else:
            raise ValueError('Unknown Iunit: {0}'.format(Iunit))
        ylabel = 'Linestrength ({0})'.format(Iunit_str)
    else:
        cutoff = 0
        try: # to find units and real name (if exists in initial databank)
            _, _, name, unit = columndescriptor[plot]
            ylabel = '{0} - {1} [{2}]'.format(name, plot, unit)
        except KeyError:
            ylabel = plot
            
    print(len(sp),'lines')
    if len(sp) == 0:
        raise ValueError('0 lines. Change your cutoff criteria?')

    # %% Plot - Plotly version

    def get_x(w):
        ''' w (input) is supposed to be in vacuum wavenumbers in the 
        lines database '''

        # Convert wavelength / wavenumber
        if wunit == 'cm-1':
            x = w
        elif wunit == 'nm':
            x = cm2nm(w)
            
            # Correct if requested medium is air        
            if medium == 'air':
                x = vacuum2air(x)
            elif medium == 'vacuum':
                pass
            else:
                raise ValueError('Unknown medium: {0}'.format(medium))
        else:
            raise ValueError('Unknown wunit: {0}'.format(wunit))
        return x

    # Parse databank to get relevant information on each line 
    # (one function per databank format)

    def get_label_hitran_locglob(row, details):
        ''' 
        Todo
        -------
        
        replace with simple astype(str) statements and str operations
        
        ex:        
        > '['+df[locl].astype(str)+']('+df[globl].astype(str)+'->'+
        >     df[globu].astype(str)'+)'
    
        will be much faster! 
        '''
        
        molecule = get_molecule(row.id)

        label = ('{0}[iso{1}]'.format(molecule, row['iso']) +
               '[{locl}]({globl})->({globu})'.format(
                **dict([(k,row[k]) for k in ['locl','globl','globu']])))
        
        for k in details:
            name, _, unit = details[k]
            if is_float(row[k]):
                label += '\n{0} {1}: {2:.3g} {3}'.format(k, name, row[k], unit)
            else:
                label += '\n{0} {1}: {2} {3}'.format(k, name, row[k], unit)
            
        return label

    def get_label_hitran_branchjglob(row, details):
        molecule = get_molecule(row.id)

        label = ('{0}[iso{1}] '.format(molecule, row['iso']) +
               '[{branch}{jl}]({globl})->({globu})'.format(
                **dict([(k,row[k]) for k in ['branch','jl','globl', 'globu']])))
        
        for k in details:
            name, _, unit = details[k]
            if is_float(row[k]):
                label += '\n{0} {1}: {2:.3g} {3}'.format(k, name, row[k], unit)
            else:
                label += '\n{0} {1}: {2} {3}'.format(k, name, row[k], unit)
            
        return label

    def get_label_hitran_branchjv(row, details):
        molecule = get_molecule(row.id)

        label = ('{0}[iso{1}]'.format(molecule, row['iso']) +
               '[{branch}{jl}]({v1l})->({v1u})'.format(
                **dict([(k,row[k]) for k in ['branch','jl','v1l', 'v1u']])))
        
        for k in details:
            name, _, unit = details[k]
            if is_float(row[k]):
                label += '\n{0} {1}: {2:.3g} {3}'.format(k, name, row[k], unit)
            else:
                label += '\n{0} {1}: {2} {3}'.format(k, name, row[k], unit)
            
        return label

    def get_label_cdsd(row, details):
        label = ('CO2[iso{iso}] [{branch}{jl}]({v1l}{v2l}`{l2l}`{v3l})->({v1u}{v2u}`{l2u}`{v3u})'.format(
                **dict([(k,row[k]) for k in ['v1u', 'v2u', 'l2u', 'v3u',
                                           'v1l', 'v2l', 'l2l', 'v3l',
                                           'branch','jl', 'iso']])))
    
        for k in details:
            name, _, unit = details[k]
            if is_float(row[k]):
                label += '\n{0} {1}: {2:.3g} {3}'.format(k, name, row[k], unit)
            else:
                label += '\n{0} {1}: {2} {3}'.format(name, k, row[k], unit)
            
        return label
    
    def get_label_none(row):
        return 'unknown databank format. \ndetails cant be read'

    # add extra info to label
    details = {}
    if columndescriptor:
        for k in lineinfo:
            if not k in sp.columns:
                raise KeyError('{0} not a {1} databank entry: {2}'.format(k,  
                               columndescriptor, dbformat.upper()))
            try: # to find units and real name (if exists in initial databank)
                _, ktype, name, unit = columndescriptor[k]
                details[k] = (name, ktype, ' [{0}]'.format(unit))
            except:
                details[k] = ('', None, '')  # keep short name

    # Get label
    if dbformat == 'hitran':  
        # Different labels for different hitran molecules
        if 'locl' in sp and 'globl' in sp and 'globu' in sp: # General case
            sp['label'] = sp.apply(lambda r: get_label_hitran_locglob(r, details), axis=1)
        elif 'branch' in sp and 'jl' in sp and 'globu' in sp and 'globl' in sp:  # ex: Group2 class 1
            sp['label'] = sp.apply(lambda r: get_label_hitran_branchjglob(r, details), axis=1)
        elif 'branch' in sp and 'jl' in sp and 'v1u' in sp and 'v1l' in sp:  # ex: Group2 class 1
            sp['label'] = sp.apply(lambda r: get_label_hitran_branchjv(r, details), axis=1)
        else:
            raise ValueError('Unknown hitran molecule format')
             
            
    elif dbformat in ['cdsd', 'cdsd-4000']:
        sp['label'] = sp.apply(lambda r: get_label_cdsd(r, details), axis=1)
    else:
        sp['label'] = sp.apply(get_label_none, axis=1)
        

    #from plotly.graph_objs import Scatter, Figure, Layout
    #
    l = [
            go.Bar(
                x=get_x(sp.shiftwav),
                y=sp[plot],
                text=sp.label,
                width=barwidth,
                name='linestrength',
                )
        ]
    
    if wunit == 'nm':
        xlabel = 'Wavelength (nm) [{0}]'.format(medium)
    elif wunit == 'cm-1':
        xlabel = 'Wavenumber (cm-1)'
    else:
        raise ValueError('unknown wunit: {0}'.format(wunit))

    if yscale == 'log':
        plot_range = (int(np.log10(max(cutoff, sp[plot].min()))),
                               int(np.log10(sp[plot].max()))+1)
    else:
        plot_range = (max(cutoff, sp[plot].min()),sp[plot].max()+1)

    layout= go.Layout(
            title= 'Line Survey ({T}K, {P:.3f}bar, Mfrac={Xi:.3f})'.format(**{'T':T, 'P':P, 'Xi':Xi}),
            hovermode= 'closest',
            xaxis= dict(
                title= xlabel,
            ),
            yaxis=dict(
                    title= ylabel,
                    #note: LaTeX doesnt seem to work in Offline mode yet.
                    type=yscale,
                    range=plot_range,
                    titlefont=dict(
                        color='#1f77b4'
                    ),
                    tickfont=dict(
                        color='#1f77b4'
                    )
                ),
            showlegend= False,
        )

    # %% Add overlay
    def add_overlay(overlay):
            
            over_w = overlay[0]
            over_I = overlay[1]
            args = overlay[2:]
            over_name = args[0] if len(args) > 0 else 'overlay' 
            over_units = args[1] if len(args) > 1 else 'a.u'
    
            l.append(go.Scatter(
                        x=over_w,
    #                        y=spec.out[overlay],
                        y=over_I,
                        yaxis='y2',
                        name=overlay,
                    ))
    
            layout['yaxis2'] = dict(
                    title= '{0} ({1})'.format(over_name.capitalize(),
                                    over_units), #note: LaTeX doesnt seem to
                                                              # work in Offline mode yet.
                    overlaying='y',
                    type='log',
                    side='right',
                    titlefont=dict(
                        color='#ff7f0e'
                    ),
                    tickfont=dict(
                        color='#ff7f0e'
                    ),
                    )
                    
                    
    if overlay is not None:
        
        if type(overlay) is not list:
            overlay = [overlay]
            
        for ov in overlay:
            add_overlay(ov)

    # %% Plot

    fig= go.Figure(data=l, layout=layout)
    py.plot(fig, filename=filename, auto_open=display)

# %% Test 
if __name__ == '__main__':

    from neq.test.spec.test_line_survey import _run_testcases
    print('Testing line survey functions:', _run_testcases(plot=True))



# -*- coding: utf-8 -*-
""" Tools to test RADIS library

Notes
-----

Created on Thu May 28 14:47:36 2015

@author: Erwan

Tools to test RADIS library

Todo
---- 

add all files in a separetaly downloadble folder. getTestFile should check
the file exist else suggest to download the folder

Examples
--------

Run all tests:
    
    cd radis/test
    pytest
    
Run only "fast" tests (tests that have "fast" in their name, and should be 
a few seconds only):
    
    cd radis/test
    pytest -k fast
    
"""

from __future__ import print_function, absolute_import, division, unicode_literals

import os
from warnings import warn
from radis.misc.config import (getDatabankList, getDatabankEntries, addDatabankEntries,
                             diffDatabankEntries)
from radis.misc.utils import FileNotFoundError
from radis.misc.basics import compare_dict, compare_paths
from os.path import join, dirname

TEST_FOLDER_PATH = join(dirname(dirname(__file__)), 'test')
IGNORE_MISSING_DATABASES = True        # TODO: move in ~/.radis

def getTestFile(file):
    ''' Return the full path of a test file. Used by test functions not to
    worry about the project architecture'''

    return join(TEST_FOLDER_PATH, 'files', file)

def getValidationCase(file):
    ''' Return the full path of a validation case file. Used by test functions not to
    worry about the project architecture'''

    return join(TEST_FOLDER_PATH, 'validation', file)

try: # Python 3.6 only 
    getTestFile.__annotations__['file'] = os.listdir(join(TEST_FOLDER_PATH, 'files'))
    getValidationCase.__annotations__['file'] = os.listdir(join(TEST_FOLDER_PATH, 'validation'))
except: 
    pass

# %% Comparison functions

def testEqual(a, b, info=''):
    if a != b:
        print('Mismatch', info, ':', a, '!=', b)
    return a == b

# %% Test Databases

TEST_DATABASES = {
        'HITRAN-CO2-TEST':{
            'info':'HITRAN 2016 database, CO2, 1 main isotope (CO2-626), bandhead: '+\
                   '2380-2398 cm-1 (4165-4200 nm)',
            'path':[getTestFile(r'hitran_co2_626_bandhead_4165_4200nm.par')],
            'format':'hitran',
            'parfuncfmt':'hapi',
            'levelsfmt':'neq', # TODO: replace with 'radis'
            }, 
        'HITRAN-CO-TEST':{
            'info':'HITRAN 2016 database, CO, 3 main isotopes (CO-26, 36, 28), '+\
                   '2000-2300 cm-1',
            'path':[getTestFile(r'hitran_co_3iso_2000_2300cm.par')],
            'format':'hitran',
            'parfuncfmt':'hapi',
            'levelsfmt':'neq', # TODO: replace with 'radis'
            },
        }

# %% Utils to test spec module

def build_test_databases(verbose=True):
    ''' Build test databases and add them in ~/.radis. Generate the file if it 
    doesnt exist
    
    In particular:
    
    - HITRAN-CO2-TEST: CO2, HITRAN 2016, 4165-4200 nm 
    - HITRAN-CO-TEST: CO, HITRAN 2016, 2000-2300 cm-1
    
    These test databases are used to run the different test routines. They can
    obviously be used by Users to run simulations, but we suggest Users to download
    their own line databases files and add them to ~/.radis so they have more control
    on it
    
    '''
    
    # Get list of databases
    try:
        dbnames = getDatabankList()
    except FileNotFoundError:
        dbnames = []
        
    # %% Add test databases
    
    def add_to_parser(config, name, dic):
        for k, v in dic.items():
            config[name][k] = v
        if verbose: print("Adding '{0}' database in ~/.radis".format(name))
        
    for dbname, dbentries in TEST_DATABASES.items():
        
        if dbname in dbnames:  # Check entries are correct
#            for k 
            diff = diffDatabankEntries(getDatabankEntries(dbname), dbentries,
                                       verbose=False)
            if diff is not None:
                raise ValueError('{0}'.format(diff)+\
                                 '\nIn ~/.radis\n----------\n{0}'.format(getDatabankEntries(dbname))+\
                                 '\n\nExpected\n---------\n{0}\n\n'.format(dbentries)+\
                                 'Test Database {0} doesnt match expected '.format(dbname)+\
                                 'entries for key `{0}`. See comparison above'.format(diff))

        else:  #  add them (create ~/.radis file if doesnt exist yet)
            addDatabankEntries(dbname, dbentries)
    
    return


# %% Deal with missing databases
def _failsafe_if_no_db(testcase, *args, **kwargs):
    '''finally not implemented?'''
    from radis.misc.utils import DatabankNotFound
    try:
        testcase(*args, **kwargs)
    except DatabankNotFound:
        import sys
        print((sys.exc_info()))
        print(('Testing {0}: Database not defined. \n'.format(testcase.__name__)+\
                       'Ignoring the test'))
        return True
    
def IgnoreMissingDatabase(err, file='', warnings=True):
    if IGNORE_MISSING_DATABASES:
        if warnings:
            import sys
            print(sys.exc_info())
            print('In {0}: Database not defined: {1}'.format(file, err.filename)+\
                  '\n Ignoring the test')
        return True
    else:
        raise err

if __name__ == '__main__':
#    run_tests()
    build_test_databases()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  2 23:39:52 2018

@author: erwan
"""

from __future__ import print_function, absolute_import, division, unicode_literals
import radis
from radis.misc.utils import DatabankNotFound
from radis.spectrum.rescale import get_redundant, get_recompute
from radis.tools.database import load_spec
from radis.test.utils import getTestFile
import numpy as np

def test_compression__fast(verbose=True, warnings=True, *args, **kwargs):
    ''' Test that redundant quantities are properly infered from already known 
    spectral quantities '''
    
    # Get spectrum
    s1 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s1.update()
    
    # Analyse redundant spectral quantities
    redundant = get_redundant(s1)
    if verbose: print(redundant)
    
    assert redundant == {'emissivity_noslit': True, 'radiance_noslit': True, 
                         'emisscoeff': True, 'transmittance_noslit': True, 
                         'absorbance': True, 'abscoeff': False}
    
    return True


def test_update_transmittance__fast(verbose=True, warnings=True, *args, **kwargs):
    ''' Test that update can correctly recompute missing quantities '''
    # TODO: add one with radiance too
    
    # Work with a Spectrum object that was generated by Specair
    s = load_spec(getTestFile('N2C_specair_380nm.spec'))
    w1, T1 = s.get('transmittance_noslit')   # our reference
    
    if verbose:
        debug_mode = radis.DEBUG_MODE  # shows all the rescale steps taken
        radis.DEBUG_MODE=True
    
    # Now let's apply some update() steps
    
    # 1) Make sure updating doesnt change anything
    s.update()
    w2, T2 = s.get('transmittance_noslit')
    
    # 2) Now recompute from abscoeff
    del s._q['transmittance_noslit']
    s.update()
    w2, T3 = s.get('transmittance_noslit')
    
    # 3) Now recompute from absorbance
    del s._q['transmittance_noslit']
    del s._q['abscoeff']
    s.update()
    w2, T4 = s.get('transmittance_noslit')

    if verbose:
        radis.DEBUG_MODE = debug_mode

    # Compare    
    assert np.allclose(T1, T2)
    assert np.allclose(T1, T3)
    assert np.allclose(T1, T4)
    
    return True

def test_get_recompute(verbose=True, *args, **kwargs):
    ''' Make sure get_recompute works as expected
    
    Here, we check which quantities are needed to recompute radiance_noslit'''

    # Equilibrium
    # -----------
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    assert s.get_vars() == ['abscoeff']
    assert s.is_at_equilibrium()
    # At equilibrium, everything should be deduced from abscoeff
    assert set(get_recompute(s, ['radiance_noslit'])) == set(('radiance_noslit', 'abscoeff'))
    
    # Non Equilibrium
    # ----------------
    s.conditions['Tvib'] = 2000 # force non equilibrium
    assert not s.is_at_equilibrium()
    # Now more data is needed:
    assert set(get_recompute(s, ['radiance_noslit'])) == set(('abscoeff', 'emisscoeff', 'radiance_noslit'))

def test_recompute_equilibrium(verbose=True, warnings=True, plot=True, 
                               *args, **kwargs):
    ''' Test that spectral quantities recomputed under equilibrium assumption
    yields the same output as with non equilibrium routines when Tvib = Trot '''
    
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
    
    # Get spectrum
    s1 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s1.rescale_path_length(100)  # just for fun
    
    assert s1.is_at_equilibrium()
    s1.update('emisscoeff')
    
    # force non equilibrium calculation by setting Tvib=2000 
    s2 = s1.copy()
    s2.conditions['Tvib'] = 2000 # force non equilibrium
    assert not s2.is_at_equilibrium()
    s2.update('radiance_noslit') 

    # update s1 now (at equilibrium)
    s1.update('radiance_noslit')
    
    s1.name = 'scaled with Kirchoff law'
    s2.name = 'scaled from emisscoeff + abscoeff with RTE'
    
    if verbose: 
        print('Checked that scaling at equilibrium with Kirchoff law yields the '+\
              'same radiance as by solving the RTE from emisscoeff and abscoeff')
    
    # Compare
    assert s1.compare_with(s2,spectra_only='radiance_noslit', plot=plot)
    

def _run_all_tests(verbose=True, warnings=True, *args, **kwargs):
    test_compression__fast(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_update_transmittance__fast(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_get_recompute(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_recompute_equilibrium(verbose=verbose, warnings=warnings, *args, **kwargs)
    
    return True

if __name__ == '__main__':
    print(('Testing test_rescale.py:', _run_all_tests(verbose=True)))


# -*- coding: utf-8 -*-
'''
HITRAN database parser 
'''

from __future__ import print_function, absolute_import, division, unicode_literals

import sys
import pandas as pd
import numpy as np
from collections import OrderedDict
import os
from os.path import exists, splitext
from six.moves import range
from six.moves import zip


# %% Hitran groups and classes
# As defined in Rothman et al, "The HITRAN 2004 molecular spectroscopic database"
# Tables 3 and 4

# Groups (define local quanta)

HITRAN_GROUP1 = ['H2O', 'O3', 'SO2', 'NO2', 'HNO3', 'H2CO', 'HOCl', 'H2O2', 'COF2',
               'H2S', 'HO2', 'HCOOH', 'ClONO2', 'HOBr', 'C2H4'] # asymmetric rotors

HITRAN_GROUP2 = ['CO2', 'N2O', 'CO', 'HF', 'HCl', 'HBr', 'HI', 'OCS', 'N2', 'HCN',
                 'C2H2', 'NO+']   # diatomic and linear molecules

HITRAN_GROUP3 = ['SF6', 'CH4']   # Spherical rotors

HITRAN_GROUP4 = ['CH3D', 'CH3Cl', 'C2H6', 'NH3', 'PH3', 'CH3OH']  # symmetric rotors

HITRAN_GROUP5 = ['O2']  # Triplet-Sigma ground electronic states

HITRAN_GROUP6 = ['NO', 'OH', 'ClO']  # Doublet-Pi ground electronic states 

# Classes (define global quanta)

HITRAN_CLASS1 = ['CO', 'HF', 'HCl', 'HBr', 'HI', 'N2', 'NO+']

HITRAN_CLASS2 = ['O2']  # Diatomic molecules with different electronic levels

HITRAN_CLASS3 = ['NO', 'OH', 'ClO']  # Diatomic molecules with doublet-Pi electronic state
    
HITRAN_CLASS4 = ['N2O', 'OCS', 'HCN']  # Linear triatomic

HITRAN_CLASS5 = ['CO2']  # Linear triatomic with large Fermi resonance

HITRAN_CLASS6 = ['H2O', 'O3', 'SO2', 'NO2', 'HOCl', 'H2S', 'HO2', 'HOBr'] # Non-linear triatomic

HITRAN_CLASS7 = ['C2H2']  # Linear tetratomic

HITRAN_CLASS8 = ['NH3', 'PH3']  # Pyramidal tetratomic

HITRAN_CLASS9 = ['H2CO', 'H2O2', 'COF2']  # Non-linear tetratomic

HITRAN_CLASS10 = ['CH4', 'CH3D', 'CH3Cl', 'C2H6', 'HNO3', 'SF6', 'HCOOH', 
                  'ClONO2', 'C2H4', 'CH3OH']  # Pentatomic or greater polyatomic

# %% Parsing functions

# General case
columns_2004 = OrderedDict([ (
               # name    # format # type  # description                                 # unit 
               'id',     ('a2',   int,   'Molecular number'                               ,''                      )),(
               'iso',    ('a1',   int,   'isotope number'                                 ,''                      )),(
               'wav',    ('a12',  float, 'vacuum wavenumber'                              ,'cm-1'                  )),(
               'int',    ('a10',  float, 'intensity at 296K'                              ,'cm-1/(molecule/cm-2)', )),(
               'A',      ('a10',  float, 'Einstein A coefficient'                         ,'s-1'                   )),(
               'airbrd', ('a5',   float, 'air-broadened half-width at 296K'               ,'cm-1.atm-1'            )),(
               'selbrd', ('a5',   float, 'self-broadened half-width at 296K'              ,'cm-1.atm-1'            )),(
               'El',     ('a10',  float, 'lower-state energy'                             ,'cm-1'                  )),(
               'Tdpair', ('a4',   float, 'temperature-dependance exponent for Gamma air'  ,''                      )),(
               'Pshft',  ('a8',   float, 'air pressure-induced line shift at 296K'        ,'cm-1.atm-1'            )),( 
               'globu',  ('a15',  str,   'electronic and vibrational global upper quanta' ,''                      )),(
               'globl',  ('a15',  str,   'electronic and vibrational global lower quanta' ,''                      )),(
               'locu',   ('a15',  str,   'electronic and vibrational local upper quanta'  ,''                      )),(
               'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
               'ierr',   ('a6',   str,   'ordered list of indices corresponding to uncertainty estimates of transition parameters'    ,''   )),(
               'iref',   ('a12',  str,   'ordered list of reference identifiers for transition parameters'                    ,''           )),(
               'lmix',   ('a1',   str,   'flag indicating the presence of additional data and code relating to line-mixing'   ,''           )),(
               'gp',     ('a7',   float, 'upper state degeneracy'                         ,''                      )),(
               'gpp',    ('a7',   float, 'lower state degeneracy'                         ,''                      ))
               ])


# quick fix # TODO: proper implementation of HITRAN classes, and groups
# Update: classes are now implemented properly, groups remain to be done 


def _generate_cache_file(fname, df):
    ''' Generate cache file for pandas Dataframe df under name fname, which 
    must be a HDF5 
    
    Parameters
    ----------
    
    fname: str
        HDF5 filename
        
    df: pandas Dataframe
        line database to store
        
    Note
    ----
    
    Performance of different compression librairies / methods:
        
    - storing as np.float32 can save up to 50% space, but we can loose some 
    information such as low linestrength numbers (1e-80, 1e-100) that becomes 0
    
    - storing with compression (complevel=9, complib='blosc') saves up to 50%
    space, but reading speed is doubled 
    
    - storing with low compression 'complevel=1', complib='blosc' can still
    save up to 50%, but reading speed is only increased by about 20%
    
    - format='fixed' is always much faster than format='table' (about a factor 5!)
    
    Eventually we use ``format='fixed', mode='w', complevel=1, complib='blosc')`` 

    
    '''
    
    assert fname.endswith('h5')
    
    df.to_hdf(fname, 'df', format='fixed', mode='w', complevel=1, complib='blosc')
#    df.to_hdf(fname, 'df', format='fixed', mode='w')  # would be 10-20% faster, but take 2x more space
    


# %% Hitran global quanta classes

def _parse_HITRAN_class1(df):
    ''' Diatomic molecules: CO, HF, HCl, HBr, HI, N2, NO+
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>      v1
    >>>  13x I2
    
    '''
    dgu = df['globu'].str.extract(
            '[ ]{13}(?P<v1u>[\d ]{2})',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{13}(?P<v1l>[\d ]{2})',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def _parse_HITRAN_class2(df):
    ''' Diatomic molecules with different electronic levels: O2
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>      X  v1
    >>>  12x A1 I2
    
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df

def _parse_HITRAN_class3(df):
    ''' Diatomic molecules with doublet-Pi electronic state: NO, OH, ClO
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -------
    
    HITRAN syntax:
    
    >>>      X i     v1
    >>>  7x A1 A3 2x I2
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df

def _parse_HITRAN_class4(df):
    ''' Parse linear triatomic class in HITRAN [1]_: N2O, OCS, HCN
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    Notes
    -----
    
    HITRAN syntax:
    
    >>>     v1 v2 l2 v3
    >>>  7x I2 I2 I2 I2
    
    Note: I2 in regexp: [\d ]{2} 
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    dgu = df['globu'].str.extract(
            '[ ]{7}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{7}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def _parse_HITRAN_class5(df):
    ''' Parse linear triatomic with large Fermi resonance in HITRAN [1]_: CO2
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    Notes
    -----
    
    HITRAN syntax:
    
    >>>     v1 v2 l2 v3 r
    >>>  6x I2 I2 I2 I2 I1
    
    Note: I2 in regexp: [\d ]{2} 
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    dgu = df['globu'].str.extract(
            '[ ]{6}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<l2u>[\d ]{2})(?P<v3u>[\d ]{2})(?P<ru>\d)',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{6}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<l2l>[\d ]{2})(?P<v3l>[\d ]{2})(?P<rl>\d)',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def _parse_HITRAN_class6(df):
    ''' Parse non-linear triatomic in HITRAN [1]_: H2O, O3, SO2, NO2, HOCl, H2S, HO2, HOBr
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    Notes
    -----
    
    HITRAN syntax:
    
    >>>     v1 v2 v3
    >>>  9x I2 I2 I2
    
    Note: I2 in regexp: [\d ]{2} 
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    dgu = df['globu'].str.extract(
            '[ ]{9}(?P<v1u>[\d ]{2})(?P<v2u>[\d ]{2})(?P<v3u>[\d ]{2})',
            expand=True)
    dgl = df['globl'].str.extract(
            '[ ]{9}(?P<v1l>[\d ]{2})(?P<v2l>[\d ]{2})(?P<v3l>[\d ]{2})',
            expand=True)
    dgu = dgu.apply(pd.to_numeric)
    dgl = dgl.apply(pd.to_numeric)
    return pd.concat([df, dgu, dgl], axis=1)

def _parse_HITRAN_class7(df):
    ''' Parse linear tetratomic in HITRAN [1]_: C2H2
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df

def _parse_HITRAN_class8(df):
    ''' Pyramidal tetratomic in HITRAN [1]_: NH3, PH3
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df

def _parse_HITRAN_class9(df):
    ''' Non-linear tetratomic in HITRAN [1]_: H2CO, H2O2, COF2
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df

def _parse_HITRAN_class10(df):
    ''' Pentatomic or greater polyatomic in HITRAN [1]_
    
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax:
    
    >>>
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    print('parse_global_quanta not implemented for molecules of this HITRAN class')
    return df
    
# %% HITRAN Local quanta
    
    
def _parse_HITRAN_group1(df):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax: [1]_
    
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df

    
    
def _parse_HITRAN_group2(df):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax: [1]
    
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    '''
    
    dgu = df['locu'].str.extract(
            '[ ]{10}(?P<Fu>.{5})',
            expand=True)
    dgl = df['locl'].str.extract(
            '[ ]{5}(?P<branch>[\S]{1})(?P<jl>[\d ]{3})(?P<sym>.)(?P<Fl>.{5})',
            expand=True)
#    dgu = dgu.apply(pd.to_numeric)
    dgl['jl'] = dgl.jl.apply(pd.to_numeric)
#    dgl = dgl.apply(pd.to_numeric)
    
    del df['locu']
    del df['locl']
    
    return pd.concat([df, dgu, dgl], axis=1)

    
#                # 10X included in Fu
#               'Fu',     ('a15',  str,  'upper state total angular momentum including nuclear spin'  ,''         )),(
#               #'locl',   ('a15',  str,   'electronic and vibrational local lower quanta'  ,''                      )),(
#               'branch', ('a6',   str,     'O, P, Q, R, S branch symbol'                  ,''                      )),(
#               'jl',     ('a3',   int,    'lower state rotational quantum number'         ,''                      )),(
#               'sym',    ('a1',   str,     'symmetry'                                     ,''                      )),(
#               'Fl',     ('a5',   str,   'lower state total angular momentum including nuclear spin', ''         )),(

    
def _parse_HITRAN_group3(df):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax [1]_:
        
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df

    
def _parse_HITRAN_group4(df):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax [1]_:
        
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df

    
def _parse_HITRAN_group5(df):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax [1]_:
        
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df

    
def _parse_HITRAN_group6(df):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
        lines read from a HITRAN-like database
        
    
    Notes
    -----
    
    HITRAN syntax [1]_:
    
        
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    '''
    print('parse_local_quanta not implemented for molecules of this HITRAN group')
    return df


# %% Reading function


def _format_dtype(dtype):
    ''' Format dtype from specific columns. Crash with hopefully helping error message '''
    
    try:
        dt = np.dtype([(str(k), c) for k, c in dtype])
        # Note: dtype names cannot be `unicode` in Python2. Hence the str()
    except TypeError:
        # Cant read database. Try to be more explicit for user
        print('Data type')
        print('-'*30)
        for (k, c) in dtype:
            print(str(k), '\t', c)
        print('-'*30)
        raise
    return dt

def _cast_to_dtype(data, dtype):
    ''' Cast array to certain type, crash with hopefull helping error message.
    Return casted data
    
    
    Parameters    
    ----------
    
    data: array to cast
    
    dtype: (ordered) list of (param, type)
    
    '''
    
    dt = _format_dtype(dtype)
    
    try:
        data = np.array(data, dtype=dt)
    except ValueError:
        try:
            # Cant read database. Try to be more explicit for user
            print('Cant cast data to specific dtype. Trying column by column:')
            print('-'*30)
            for i in range(len(data[0])):
                print(dtype[i], '\t', np.array(data[0][i], dtype=dt[i]))
            print('-'*30)
        except ValueError:
            print('>>> Next param:', dtype[i], '. Value:', data[0][i], '\n')
            raise ValueError('Cant cast data to specific dtype. Tried column by column. See results above')

    return data

def hit2df(fname, count=-1, cache=False, verbose=True):
    ''' Convert a HITRAN/HITEMP [1]_ file to a Pandas dataframe 
    
    
    Parameters    
    ----------
    
    fname: str
        HITRAN-HITEMP file name 
        
    count: int
        number of items to read (-1 means all file)
        
    cache: boolean
        if True, a pandas-readable HDF5 file is generated on first access, 
        and later used. This saves on the datatype cast and conversion and
        improves performances a lot (but changes in the database are not 
        taken into account). If False, no database is used. If 'regen', temp
        file are reconstructed. Default False. 
    
    
    Returns
    -------
    
    df: pandas Dataframe
        dataframe containing all lines and parameters
        
    
    
    References
    ----------

    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    
    
    Notes
    -----
    
    Performances: see CDSD-HITEMP parser
    
    '''
    
    columns = columns_2004

    if cache: # lookup if cached file exist. 
#        fcache = fname+'.cached'
        fcache = splitext(fname)[0]+'.h5'
        if exists(fcache):
            if cache == 'regen':
                os.remove(fcache)
                if verbose: print('Deleted h5 cache file : {0}'.format(fcache))
            else:
                if verbose: print('Using h5 file: {0}'.format(fcache))
    #            return pd.read_csv(fcache)
                return pd.read_hdf(fcache, 'df')
        
    # Detect the molecule by reading the start of the file
    with open(fname) as f:
        mol = get_molecule(int(f.read(2)))
        
    # %% Start reading the full file
    
    # To be faster, we read file totally in bytes mode with fromfiles. But that
    # requires to properly decode the line return character:
    
    # get format of line return
    if sys.platform in ['linux', 'darwin', 'linux2']:
        linereturnformat = 'a1'
    elif sys.platform in ['win32']:
        linereturnformat = 'a2'
    else:
        raise ValueError('Line return format not defined for this OS: please update RADIS')

    # ... Create a dtype with the binary data format and the desired column names
    dtype = [(k, c[0]) for (k, c) in columns.items()]+[('_linereturn',linereturnformat)]   
    # ... _linereturn is to capture the line return symbol. We delete it afterwards
    dt = _format_dtype(dtype)
    data = np.fromfile(fname, dtype=dt, count=count)
    
    # ... Cast to new type
    # This requires to recast all the data already read, but is still the fastest
    # method I found to read a file directly (for performance benchmark see 
    # CDSD-HITEMP parser)
    newtype = [c[0] if (c[1]==str) else c[1] for c in columns.values()]
    dtype = list(zip(list(columns.keys()), newtype))+[('_linereturn',linereturnformat)]
    data = _cast_to_dtype(data, dtype)

    # %% Create dataframe    
    df = pd.DataFrame(data.tolist(), columns=list(columns.keys())+['_linereturn'])

    # assert one molecule per database only. Else the groupbase data reading 
    # above doesnt make sense
    nmol = len(set(df['id']))
    if nmol == 0:
        raise ValueError('Databank looks empty')        
    elif nmol!=1:
        # Crash, give explicity error messages
        try:
            secondline = df.iloc[1]
        except IndexError:
            secondline = ''
        raise ValueError('Multiple molecules in database ({0}). Current '.format(nmol)+\
                         'spectral code only computes 1 species at the time. Use MergeSlabs. '+\
                         'Verify the parsing was correct by looking at the first row below: '+\
                         '\n{0}'.format(df.iloc[0])+'\n----------------\nand the second row '+\
                         'below: \n{0}'.format(secondline))
    
    for k, c in columns.items():
        if c[1] == str:
            df[k] = df[k].str.decode("utf-8")
    
    # %% Add local quanta attributes, based on the HITRAN group
    df = parse_local_quanta(df, mol)
    
    # %% Add global quanta attributes, based on the HITRAN class
    df = parse_global_quanta(df, mol)

    # Strip whitespaces around PQR columns (due to 2 columns jumped)
    if 'branch' in df:
        df['branch'] = df.branch.str.strip()
    
    # Delete dummy column than handled the line return character
    del df['_linereturn']
    
    if cache: # cached file mode but cached file doesn't exist yet (else we had returned)
        if verbose: print('Generating cached file: {0}'.format(fcache))
        try:
#            df.to_csv(fcache)
            _generate_cache_file(fcache, df)
        except:
            if verbose:
                print(sys.exc_info())
                print('An error occured in cache file generation. Lookup access rights')
            pass
        
    return df 

def parse_local_quanta(df, mol):
    '''
    Parameters
    ----------
    
    df: pandas Dataframe
    
    mol: str
        molecule name
    '''
    
    # had some problems with bytes types
    df['locu'] = df.locu.astype(str)
    df['locl'] = df.locl.astype(str)
    
    if mol in HITRAN_GROUP1:
        df = _parse_HITRAN_group1(df)
    elif mol in HITRAN_GROUP2:
        df = _parse_HITRAN_group2(df)
    elif mol in HITRAN_GROUP3:
        df = _parse_HITRAN_group3(df)
    elif mol in HITRAN_GROUP4:
        df = _parse_HITRAN_group4(df)
    elif mol in HITRAN_GROUP5:
        df = _parse_HITRAN_group5(df)
    elif mol in HITRAN_GROUP6:
        df = _parse_HITRAN_group6(df)
    else:
        raise ValueError('Unknown group for molecule {0}. Cant parse local quanta'.format(
                mol))
        
    return df
            
def parse_global_quanta(df, mol):
    ''' 
    
    Parameters
    ----------
    
    df: pandas Dataframe
    
    mol: str
        molecule name
    '''
    
    # had some problems with bytes types
    df['globu'] = df.globu.astype(str)
    df['globl'] = df.globl.astype(str)
    
    if mol in HITRAN_CLASS1:
        df = _parse_HITRAN_class1(df)
    elif mol in HITRAN_CLASS2:
        df = _parse_HITRAN_class2(df)
    elif mol in HITRAN_CLASS3:
        df = _parse_HITRAN_class3(df)
    elif mol in HITRAN_CLASS4:
        df = _parse_HITRAN_class4(df)
    elif mol in HITRAN_CLASS5:
        df = _parse_HITRAN_class5(df)
    elif mol in HITRAN_CLASS6:
        df = _parse_HITRAN_class6(df)
    elif mol in HITRAN_CLASS7:
        df = _parse_HITRAN_class7(df)
    elif mol in HITRAN_CLASS8:
        df = _parse_HITRAN_class8(df)
    elif mol in HITRAN_CLASS9:
        df = _parse_HITRAN_class9(df)
    elif mol in HITRAN_CLASS10:
        df = _parse_HITRAN_class10(df)
    else:
        raise ValueError('Unknown class for molecule {0}. Cant parse global quanta'.format(
                mol))
        
    return df
            

def get_molecule_identifier(molecule_name):
    '''
    For a given input molecular formula, return the corresponding HITRAN molecule 
    identifier number [1]_.
    
    
    Parameters
    ----------
    molecular_formula : str
        The string describing the molecule.
        
        
    Returns
    -------
    M : int
        The HITRAN molecular identified number.
        
        
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    Function is from from https://github.com/nzhagen/hitran/blob/master/hitran.py
    
    '''

    trans = { '1':'H2O',    '2':'CO2',   '3':'O3',      '4':'N2O',   
              '5':'CO',    '6':'CH4',   '7':'O2',     '8':'NO',
              '9':'SO2',   '10':'NO2',  '11':'NH3',    '12':'HNO3', 
              '13':'OH',   '14':'HF',   '15':'HCl',   '16':'HBr',
             '17':'HI',    '18':'ClO',  '19':'OCS',    '20':'H2CO', 
             '21':'HOCl', '22':'N2',   '23':'HCN',   '24':'CH3Cl',
             '25':'H2O2',  '26':'C2H2', '27':'C2H6',   '28':'PH3',  
             '29':'COF2', '30':'SF6',  '31':'H2S',   '32':'HCOOH',
             '33':'HO2',   '34':'O',    '35':'ClONO2', '36':'NO+',  
             '37':'HOBr', '38':'C2H4', '39':'CH3OH', '40':'CH3Br',
             '41':'CH3CN', '42':'CF4',  '43':'C4H2',   '44':'HC3N', 
             '45':'H2',   '46':'CS',   '47':'SO3'}
    ## Invert the dictionary.
    trans = {v:k for k,v in trans.items()}

    return int(trans[molecule_name])

def get_molecule(molecule_id):
    '''
    For a given input molecular identifier, return the corresponding HITRAN 
    molecule name [1]_.
    
    
    Parameters    
    ----------
    
    molecular_id : str
        Hitran identifier of the molecule.
        
    
    References
    ----------
    
    .. [1] `HITRAN 1996, Rothman et al., 1998 <https://www.sciencedirect.com/science/article/pii/S0022407398000788>`__
    
    '''
    
    # assert str
    id = '{0}'.format(molecule_id)

    trans = { '1':'H2O',    '2':'CO2',   '3':'O3',      '4':'N2O',   
              '5':'CO',    '6':'CH4',   '7':'O2',     '8':'NO',
              '9':'SO2',   '10':'NO2',  '11':'NH3',    '12':'HNO3', 
              '13':'OH',   '14':'HF',   '15':'HCl',   '16':'HBr',
             '17':'HI',    '18':'ClO',  '19':'OCS',    '20':'H2CO', 
             '21':'HOCl', '22':'N2',   '23':'HCN',   '24':'CH3Cl',
             '25':'H2O2',  '26':'C2H2', '27':'C2H6',   '28':'PH3',  
             '29':'COF2', '30':'SF6',  '31':'H2S',   '32':'HCOOH',
             '33':'HO2',   '34':'O',    '35':'ClONO2', '36':'NO+',  
             '37':'HOBr', '38':'C2H4', '39':'CH3OH', '40':'CH3Br',
             '41':'CH3CN', '42':'CF4',  '43':'C4H2',   '44':'HC3N', 
             '45':'H2',   '46':'CS',   '47':'SO3'}

    return trans[id]

## ======================================================
# %% Test

if __name__ == '__main__':
    from radis.test.test_io import test_hitran_parser__fast
    print('Testing HITRAN parsing: ', test_hitran_parser__fast())
        


# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 14:56:25 2017

@author: erwan

Runs tests for neq/phys so that they can be accessed by pytest (and hopefully 
the CI test suite)

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have 'fast' in their name)

>>> pytest -k fast

"""



from __future__ import absolute_import, unicode_literals, print_function

import numpy as np
from numpy import isclose
from radis.phys.convert import (J2eV, J2cm, cm2J, eV2cm, eV2K, eV2nm, nm2eV,
                                dnm2dcm, nm2cm, dnm2dhz, dhz2dnm, cm2eV, eV2J, K2J,
                                J2K, dcm2dnm, K2eV, hz2nm, nm2hz,
                                torr2bar, torr2atm, bar2torr, bar2atm, atm2torr,
                                atm2bar)
from radis.phys.units import uarray, Q_, conv2

def test_convert__fast(verbose=True, *args, **kwargs):
    ''' Test conversions  '''
    
    E = np.linspace(1, 5, 5)  # eV
    assert (J2eV(K2J(J2K(eV2J(E)))) == E).all()

    E = 2150  # cm-1
    assert J2cm(cm2J(E))

    E = 1  # eV
    assert (eV2cm(E) == J2cm(eV2J(1)))
    assert (round(eV2K(E), 0) == 11605)
    assert K2eV(eV2K(E)) == E
    
    E = 250 # nm
    assert isclose(nm2eV(E),cm2eV(nm2cm(E)))
    assert (eV2nm(nm2eV(E))==E)
    assert hz2nm(nm2hz(E)) == E

    fwhm = 1.5 # nm
    lbd_0 = 632.8 # nm
    assert (isclose(fwhm, dcm2dnm(dnm2dcm(fwhm, lbd_0), nm2cm(lbd_0))))

    fwhm = 2e-3  # 0.002 nm
    lbd_0 = 632.8
    fwhm_hz = dnm2dhz(fwhm, lbd_0)         #  ~ 1.5 GHz
    if verbose: print(('{0:.2g} nm broadening at {1} nm = {2:.2g} Ghz'.format(fwhm, lbd_0, fwhm_hz*1e-9)))
    assert isclose(fwhm_hz*1e-9, 1.4973307983125002)
    assert isclose(dhz2dnm(fwhm_hz, nm2hz(lbd_0)), fwhm)
    
    assert atm2bar(1) == 1.01325
    assert isclose(torr2bar(atm2torr(bar2atm(1))), 1)
    assert isclose(torr2atm(bar2torr(atm2bar(1))), 1)


    return True

def test_units__fast(verbose=True, *args, **kwargs):

    # Test unit-ware arrays
    a = uarray(np.linspace(10, 100, 10), 'Td')          # RADIS pint-aware array
    res = Q_(np.linspace(1e-16, 1e-15, 10), 'V * cm^2')  # pint definition

    assert (np.round(np.array(a.to('V * cm^2')) - np.array(res), 5)
         == np.zeros_like(res)).all()
#    b = (print(a.to('V * cm^2'))==print(res))


    # Test conversion 
    convtable = [
            (500, 'nm', 0.5, 'µm'),
            (1, 'erg/s', 1e-7, 'W'),
            (1, 'm2',10000, 'cm2')
            ]
    for a, f, r, t in convtable:
        cr = conv2(a, f, t)
        if verbose: print(('{0} {1} = {2} {3}'.format(a, f, cr, t)))
        assert isclose(cr,r)

    return True

def _run_testcases(*args, **kwargs):
    
    assert test_convert__fast(*args, **kwargs)
    assert test_units__fast(*args, **kwargs)
    
    return True

if __name__== '__main__':
    print('testing phys.py:', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have 'fast' in their name)

>>> pytest -k fast

"""

from __future__ import print_function, absolute_import, division, unicode_literals
import numpy as np
from radis.test.utils import getTestFile
from radis.tools.database import load_spec
from radis.spectrum.compare import get_distance, plot_diff

# Test routines

def test_compare_methods(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Just run all Spectrum compare methods to check they work'''

    if plot and close_plots:
        import matplotlib.pyplot as plt
        plt.close('all')
    
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s.resample(np.linspace(2193, 2193.8, 100))   # limits to a single line, because get_distance() 
                                                   # is very computationaly heavy 
    s.update('radiance_noslit')
    s_noabsorption = s.copy()
    s.name = 'solve RTE'
    s_noabsorption.name = 'optically thin'
    
    # rescale, normal 
#    s.rescale_mole_fraction(10)
    s.rescale_path_length(10)
    
    # rescale, optically thin mode
    s_noabsorption.conditions['self_absorption'] = False
#    s_noabsorption.rescale_mole_fraction(10)
    s_noabsorption.rescale_path_length(10)
    
    # Compare 
    get_distance(s, s_noabsorption, 'radiance_noslit')  # should be added in an example with experimental fit of bandhead
    title = 'CO x={0:.2f}, L={1:.2f}cm'.format(s.conditions['mole_fraction'], s.conditions['path_length'])
    if plot:
        plot_diff(s, s_noabsorption, method='diff', title=title)
        plot_diff(s, s_noabsorption, method='ratio', normalize=True, title=title)
    
    
    
# %%

def _run_testcases(plot=True, verbose=True, warnings=True, *args, **kwargs):
    ''' Test procedures
    '''
    
    # Test all Spectrum compare methods 
    # ----------------------------------
    test_compare_methods(verbose=verbose, plot=plot, *args, **kwargs)
    return True


if __name__ == '__main__':
    print('Test_compare.py: ', _run_testcases())



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 10:18:48 2017

@author: erwan

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have 'fast' in their name)

>>> pytest -k fast

"""

from __future__ import print_function, absolute_import, division, unicode_literals
#from neq.spec import SpectrumFactory
from radis.spectrum.spectrum import calculated_spectrum, transmittance_spectrum
from radis.tools.database import load_spec
from radis.los.slabs import SerialSlabs
from radis.tools.slit import (gaussian_slit, triangular_slit, trapezoidal_slit,
                           import_experimental_slit, convolve_with_slit, 
                           get_FWHM, get_effective_FWHM)
from radis.phys.units import is_homogeneous
from radis.phys.convert import dcm2dnm, dnm2dcm
#from radis.misc.utils import DatabankNotFound
#from radis.test.utils import IgnoreMissingDatabase, build_test_databases
import matplotlib.pyplot as plt 
import numpy as np
from numpy import sqrt, linspace, abs, trapz
from os.path import basename

fig_prefix = basename(__file__)+': '

# %% --------------------------------------------------------------------------
#  Test cases
# -----------------------------------------------------------------------------

def test_all_slit_shapes__fast(FWHM=0.4, verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Test all slit generation functions and make sure we get the expected FWHM'''

    if plot:
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
        if close_plots: plt.close('all')
    
    # get spectrum
    from radis.test.utils import getTestFile
    from radis.spectrum.spectrum import Spectrum
    s = Spectrum.from_txt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000.txt'),
                          quantity='radiance_noslit', waveunit='nm', unit='mW/cm2/sr/µm',
                          conditions={'medium':'air'})
    wstep = np.diff(s.get_wavelength())[0]
    
    # Plot all slits
    # ... gaussian
    s.apply_slit(FWHM, unit='nm', shape='gaussian', plot_slit=plot)
    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=2*wstep)
    
    # ... triangular
    s.apply_slit(FWHM, unit='nm', shape='triangular', plot_slit=plot)
    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=2*wstep)
    
    # ... trapezoidal
    s.apply_slit((FWHM*0.9, FWHM*1.1), unit='nm', shape='trapezoidal', plot_slit=plot)
    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=2*wstep)
    
#    # ... trapezoidal
#    s.apply_slit(FWHM, unit='nm', shape='trapezoidal', plot_slit=plot, norm='max')
#    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=1.1*wstep)
    
    # ... experimental
    s.apply_slit(getTestFile('slitfunction.txt'), unit='nm', plot_slit=plot)
    assert np.isclose(get_effective_FWHM(*s.get_slit()), FWHM, atol=0.01)
    # note that we're applying a slit function measured at 632.5 nm to a Spectrum 
    # at 4.7 µm. It's just good for testing the functions
    
#    # ... experimental, convolve with max
#    s.apply_slit(getTestFile('slitfunction.txt'), unit='nm', norm_by='max', plot_slit=plot)
#    assert np.isclose(get_FWHM(*s.get_slit()), FWHM, atol=1.1*wstep)
    
    if verbose: print('\n>>> _test_all_slits yield correct FWHM (+- wstep) : OK\n')
    
    return True # nothing defined yet

def test_slit_unit_conversions_spectrum_in_cm(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Test that slit is consistently applied for different units
    
    Assert that:
    
    - calculated FWHM is the one that was applied
    
    '''
    
    from radis.test.utils import getTestFile
    from radis.tools.database import load_spec
    
    if plot: 
        plt.ion()
        if close_plots: plt.close('all')
    
    # %% Get a Spectrum (stored in cm-1)
    s_cm = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    s_cm.rescale_mole_fraction(1)   # just because it makes better units
    s_cm.update()
    wstep = s_cm.conditions['wstep']
    
    assert s_cm.get_waveunit() == 'cm-1'       # ensures it's stored in cm-1
    
    for shape in ['gaussian', 'triangular']:
        
        # Apply slit in cm-1 
        slit_cm = 2    
        s_cm.name = 'Spec in cm-1, slit {0:.2f} cm-1'.format(slit_cm)
        s_cm.apply_slit(slit_cm, unit='cm-1', shape=shape, mode='same')  
        # ... mode=same to keep same output length. It helps compare both Spectra afterwards
        fwhm = get_FWHM(*s_cm.get_slit())   # in cm-1 as that's s.get_waveunit()
        assert np.isclose(slit_cm, fwhm, atol=2*wstep)
            
        # Apply slit in nm this time
        s_nm = s_cm.copy()
        w_cm = s_nm.get_wavenumber(which='non_convoluted')
        slit_nm = dcm2dnm(slit_cm, w_cm[len(w_cm)//2])
        s_nm.name = 'Spec in cm-1, slit {0:.2f} nm'.format(slit_nm)
        s_nm.apply_slit(slit_nm, unit='nm', shape=shape, mode='same')
        
        plotargs = {}
        if plot: plotargs['title'] = 'test_slit_unit_conversions: {0} ({1} cm-1)'.format(shape, slit_cm)
        s_cm.compare_with(s_nm, spectra_only='radiance', rtol=1e-3, 
                          verbose=verbose, plot=plot, **plotargs)
    
def test_slit_unit_conversions_spectrum_in_nm(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Test that slit is consistently applied for different units
    
    Assert that:
    
    - calculated FWHM is the one that was applied
    
    '''
    
    from radis.test.utils import getTestFile
    from radis.spectrum.spectrum import Spectrum
 
    
    if plot: 
        plt.ion()
        if close_plots: plt.close('all')
   
    # %% Get a Spectrum (stored in nm)
    s_nm = Spectrum.from_txt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000.txt'),
                          quantity='radiance_noslit', waveunit='nm', unit='mW/cm2/sr/µm',
                          conditions={'medium':'air',
                                      'self_absorption':False})
    s_nm.rescale_path_length(1, 0.001)   # just because it makes better units
    wstep = np.diff(s_nm.get_wavelength())[0]
    
    assert s_nm.get_waveunit() == 'nm'       # ensures it's stored in cm-1
    
    for shape in ['gaussian', 'triangular']:
        
        # Apply slit in nm
        slit_nm = 0.5    
        s_nm.name = 'Spec in nm, slit {0:.2f} nm'.format(slit_nm)
        s_nm.apply_slit(slit_nm, unit='nm', shape=shape, mode='same')  
        # ... mode=same to keep same output length. It helps compare both Spectra afterwards
        fwhm = get_FWHM(*s_nm.get_slit())   # in cm-1 as that's s.get_waveunit()
        assert np.isclose(slit_nm, fwhm, atol=2*wstep)
    
        # Apply slit in nm this time
        s_cm = s_nm.copy()
        w_nm = s_nm.get_wavelength(which='non_convoluted')
        slit_cm = dnm2dcm(slit_nm, w_nm[len(w_nm)//2])
        s_cm.name = 'Spec in nm, slit {0:.2f} cm-1'.format(slit_cm)
        s_cm.apply_slit(slit_cm, unit='cm-1', shape=shape, mode='same')
        
        plotargs = {}
        if plot: plotargs['title'] = 'test_slit_unit_conversions: {0} ({1} nm)'.format(shape, slit_nm)
        s_nm.compare_with(s_cm, spectra_only='radiance', rtol=1e-3, 
                          verbose=verbose, plot=plot, **plotargs)
    
    
    # %% 
    
def test_against_specair_convolution__fast(plot=True, close_plots=True, verbose=True, debug=False, *args, **kwargs):

    if plot:
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
        if close_plots: plt.close('all')
    
    # Test
    from radis.test.utils import getTestFile
    
    # Plot calculated vs convolved with slit
    w, I = np.loadtxt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000.txt')).T   # Specair units: mW/cm2/sr/µm
    s = calculated_spectrum(w, I, conditions={'Tvib':3000, 'Trot':1200, 
                                              'medium':'air'},
                             Iunit='mW/cm2/sr/µm')

    if plot:
        fig = plt.figure(fig_prefix+'SPECAIR convoluted vs SPECAIR non convoluted')
        s.plot('radiance_noslit', nfig=fig.number, label='calc')
    slit_nm = 0.1
    s.apply_slit(slit_nm, shape='triangular', norm_by='area')
    if plot:
        s.plot('radiance', nfig=fig.number, label='slit {0}nm'.format(slit_nm),
               color='r', lw=2)
        plt.legend()

    # Compare with Specair slit function
    s.apply_slit(slit_nm, norm_by='max')
    # Note unit conversion from mW/cm2/sr/um*nm is properly done!
    if plot:
        fig = plt.figure(fig_prefix+'convoluted RADIS vs convoluted SPECAIR')
        s.plot('radiance', nfig=fig.number, label='slit {0}nm, norm with max'.format(slit_nm),
               color='r', lw=2, Iunit='mW/cm2/sr')
    # ... Get initial spectrum convoluted with Specair
    ws, Is = np.loadtxt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000_slit0.1.txt')).T  # Specair units: mW/cm2/sr
    if plot:
        plt.plot(ws, Is, 'k', label='normalized in Specair (sides not cropped)')

    # ... Test output is the same
    wu, Iu = s.get('radiance', Iunit='mW/cm2/sr')

    As = np.trapz(Is, ws)  # Todo one day: replace with get_power() function
    Au = np.trapz(Iu, wu)
    if verbose:
        print(('Integrals should match: {0:.2f} vs {1:.2f} ({2:.2f}% error)'.format(
                As, Au, 100*abs(As-Au)/As)))
    assert np.isclose(As, Au, rtol=1e-2)

    # Test resampling
    s.resample(linspace(376,380.6,3000), unit='nm', if_conflict_drop='convoluted')
    s.apply_slit(slit_nm, norm_by='max')
    if plot:
        s.plot('radiance', Iunit='mW/cm2/sr', lw=3, nfig=fig.number, color='b',
               zorder=-3, label='Resampled')
        plt.legend()

    if verbose: print('\n>>>Testing spectrum slit matches Specair: OK')
    
    return True

def test_normalisation_mode__fast(plot=True, close_plots=True, verbose=True, *args, **kwargs):
    ''' Test norm_by = 'area' vs norm_by = 'max' '''

    from radis.test.utils import getTestFile
    
    if plot:
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
        if close_plots: plt.close('all')
    
        from publib import set_style
        set_style('origin')
    
    # %% Compare spectra convolved with area=1 and max=1
    # Slit in nm
    # Spectrum in nm
    w, I = np.loadtxt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000.txt')).T   # Specair units: mW/cm2/sr/µm
    s = calculated_spectrum(w, I, conditions={'Tvib':3000, 'Trot':1200},
                             Iunit='mW/cm2/sr/µm')
    
    FWHM = 2
    
    s.apply_slit(FWHM, norm_by='area')   # spectrum convolved with area=1
    w_area, I_area = s.get('radiance')
    if plot:
        fig = plt.figure(fig_prefix+'Spectrum in nm + slit in nm')
        fig.clear()
        ax = fig.gca()
        s.plot(nfig=fig.number, wunit='nm', label='norm_by: area', lw=3)
    s.apply_slit(FWHM, norm_by='max')    # spectrum convolved with max=1     
    w_max, I_max = s.get('radiance', wunit='nm')
    if plot:
        ax.plot(w_max, I_max/FWHM, 'r', label='(norm_by:max)/FWHM')
        ax.legend(loc='best')
    assert np.allclose(I_area, I_max/FWHM)
    if verbose: 
        print("equivalence of normalisation mode for spectrum in 'nm': OK") 
    
    # %% Compare spectra convolved with area=1 and max=1
    # Slit in nm
    # Spectrum in cm-1
    
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s.update()
    s.apply_slit(FWHM, norm_by='area', plot_slit=plot)   # spectrum convolved with area=1
    w_area, I_area = s.get('radiance')
    if plot:
        fig = plt.figure(fig_prefix+'Spectrum in cm-1 + slit in nm')
        fig.clear()
        ax = fig.gca()
        s.plot(nfig=fig.number, wunit='nm', label='norm_by: area', lw=3)
    s.apply_slit(FWHM, norm_by='max', plot_slit=plot)    # spectrum convolved with max=1
    w_max, I_max = s.get('radiance', wunit='nm')
    if plot:
        ax.plot(w_max, I_max/FWHM, 'r', label='(norm_by:max)/FWHM')
        ax.legend(loc='best')
    assert np.allclose(I_area, I_max/FWHM)
    if verbose: 
        print("equivalence of normalisation mode for spectrum in 'cm-1': {0}: OK")
    assert is_homogeneous(s.units['radiance'], 'mW/cm2/sr')
    if verbose: 
        print(("radiance unit ({0}) is homogeneous to 'mW/cm2/sr': OK".format(s.units['radiance'])))
    
    return True


def test_slit_energy_conservation__fast(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' Convoluted and non convoluted quantities should have the same area
    (difference arises from side effects if the initial spectrum is not 0 on 
    the sides '''
    
    from radis.test.utils import getTestFile
    
    if plot:
        import matplotlib.pyplot as plt
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
    if close_plots: plt.close('all')
    
    if verbose: print('\n>>> _test_slit_energy_conservation\n')

    s = calculated_spectrum(*np.loadtxt(getTestFile('calc_N2C_spectrum_Trot1200_Tvib3000_slit0.1.txt')).T, 
                            wunit='nm', Iunit='mW/cm2/sr/nm')   # arbitrary)
    
    P = s.get_power(unit='mW/cm2/sr')
    s.apply_slit(1.5, norm_by='area')
    w, I = s.get('radiance', wunit='nm', Iunit='mW/cm2/sr/nm')
    Pc = abs(np.trapz(I, x=w))  # mW/cm2/sr
    
    b = np.isclose(P,Pc, 3e-2)   

    if plot:
        fig = plt.figure(fig_prefix+'energy conservation during resampling')
        s.plot(nfig=fig.number, label='{0:.1f} mW/cm2/sr'.format(P))
        s.plot('radiance_noslit', nfig=fig.number, label='{0:.1f} mW/cm2/sr'.format(Pc))
        plt.title('Energy conservation: {0}'.format(b)) 
        plt.legend()
        plt.tight_layout()
        
    assert np.isclose(P,Pc, 3e-2)   
        
    return True
    

def test_slit_function_effect__fast(verbose=True, plot=True, close_plots=True, *args, **kwargs):
    ''' A test case to show the effect of wavelength dispersion (cf spectrometer
    reciprocal function) on the slit function '''

    from radis.test.utils import getTestFile
    from publib import fix_style
    from numpy import pi, tan, cos

    if plot:
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest
        if close_plots: plt.close('all')
    
    # Effect of Slit dispersion

    def dirac(w0, width=20, wstep=0.009):
        ''' Return a Dirac in w0. Plus some space on the side '''

        w = np.arange(w0-width, w0+width+wstep, wstep)
        I = np.zeros_like(w)
        I[len(I)//2] = 1/wstep

        return w, I

    def linear_dispersion(w, f=750, phi=-6, m=1, gr=300):
        ''' dlambda / dx
        Default values correspond to Acton 750i

        Input
        -------
        f: focal length (mm)
             default 750 (SpectraPro 2750i)

        phi: angle in degrees (°)
            default 9

        m: order of dispersion
            default 1

        gr: grooves spacing (gr/mm)
            default 300
        '''
        # correct units:
        phi *= 2*pi/360
        d = 1e-3/gr
        disp = w/(2*f)*(tan(phi)+sqrt((2*d/m/(w*1e-9)*cos(phi))**2-1))
        return disp  # to nm/mm

    w_slit, I_slit = import_experimental_slit(getTestFile('slitfunction.txt'))

    if plot:
        plt.figure(fig_prefix+'Linear dispersion effect')
        plt.plot(w_slit, I_slit, '--k', label='Exp: FWHM @{0}nm: {1:.3f} nm'.format(632.8,
                 get_effective_FWHM(w_slit, I_slit)))
        
    # Test how slit function FWHM scales with linear_dispersion
    for w0, FWHM in zip([380, 1000, 4200, 5500],
                        [0.396, 0.388, 0.282, 0.188]):
        w, I = dirac(w0)
        wc, Ic = convolve_with_slit(w, I, w_slit, I_slit, norm_by='area', 
                                    slit_dispersion=linear_dispersion)
        assert np.isclose(FWHM, get_effective_FWHM(wc, Ic), atol=0.001)

        if plot: 
            plt.plot(wc, Ic, label='FWHM @{0:.2f} nm: {1:.3f} nm'.format(w0, 
                     get_effective_FWHM(wc, Ic)))
    
    if plot:
        plt.xlabel('Wavelength (nm)')
        plt.ylabel('Dirac $x$ slit function')
        plt.legend(loc='best', prop={'size':15})
        fix_style('article')

    return True # nothing defined yet

def _run_testcases(plot=True, close_plots=False, verbose=True, *args, **kwargs):

    test_against_specair_convolution__fast(plot=plot, close_plots=close_plots, verbose=verbose,
                                           *args, **kwargs)
    test_normalisation_mode__fast(plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs)
    test_slit_energy_conservation__fast(plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs)
    test_slit_function_effect__fast(plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs)
#    test_constant_source(plot=plot, verbose=verbose, *args, **kwargs)
    test_all_slit_shapes__fast(plot=plot, close_plots=close_plots, verbose=verbose, *args, **kwargs)
    test_slit_unit_conversions_spectrum_in_cm(verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs)
    test_slit_unit_conversions_spectrum_in_nm(verbose=verbose, plot=plot, close_plots=close_plots, *args, **kwargs)
#    test_resampling__fast(plot=plot, verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    print(('Testing slit.py: ', _run_testcases(plot=True)))
    



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 15 14:56:25 2017

@author: erwan

Runs tests for neq/misc so that they can be accessed by pytest (and hopefully 
the CI test suite)

Use
------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have 'fast' in their name)

>>> pytest -k fast

"""



from __future__ import absolute_import

from __future__ import print_function
def test_config__fast(*args, **kwargs):
    from radis.misc.config import _test
    return _test(*args, **kwargs)
    
def test_utils__fast(*args, **kwargs):
    from radis.misc.utils import _test
    return _test(*args, **kwargs)
    

def _run_testcases(verbose=True, *args, **kwargs):

    test_config__fast(verbose=verbose,*args, **kwargs)
    test_utils__fast(verbose=verbose,*args, **kwargs)
    
    return True

if __name__ == '__main__':
    print(('Testing misc.py: ', _run_testcases(verbose=True)))
    